/* Pure Data: Copyright (c) 1997-1999 Miller Puckette. 

-----------------------------------------------------------------------------------------------------------------------

d_doubletest.c   Katja Vetter 2011.
 
d_doubletest.c proposes alternatives for PD classes phasor~, cos~, osc~, and vcf~,
prepared for optional single / double precision builds of Pd. It defines classes 
phasor2~, cos2~, osc2~, and vcf2~, with the purpose to test these for precision 
and performance. No type punning optimizations are implemented. 

Classes bigorsmall_pun~ and bigorsmall_nopun~ are added to compare possible
definitions for PD_BIGORSMALL in the API m_pd.h. 

*/


#include "m_pd.h"
#include <math.h>
#include <float.h>		// <float.h> defines machine epsilon
#include <stdint.h>

#define TWOPI 6.283185307179586
#define COSTABMASK (COSTABSIZE-1)

// find machine epsilon (largest relative rounding error) for t_float 
#if PD_FLOAT_PRECISION == 32
#define EPSILON FLT_EPSILON
#elif PD_FLOAT_PRECISION == 64
#define EPSILON DBL_EPSILON
#endif



/*------------------------------------------------ global cosine table ---------------------------------*/

t_float *cos_table2;		// global pointer to cosine table

static void cos_maketable2(void)
{
	if(cos_table2) return;
	cos_table2 = (t_float *)getbytes(sizeof(t_float) * (COSTABSIZE+1));
	
	t_float *costab = cos_table2;				
	t_float angle =  TWOPI / COSTABSIZE;
	int tabphase = 0, n;
	
	for(n=0; n<(COSTABSIZE+2); n++) *costab++ = cos(angle * (n & COSTABMASK));
}


/* ----------------------------------------------- phasor2~ ------------------------------------------- */

static t_class *phasor2_class;

typedef struct
{
	t_object x_obj;
	double phase;
	t_float oneoversamplerate;
	t_float f;						// scalar frequency 
} t_phasor2;


static void *phasor2_new(t_floatarg f)
{
	t_phasor2 *x = (t_phasor2 *)pd_new(phasor2_class);
	inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_float, gensym("ft1"));
	x->f = f;
	x->phase = 0.;
	x->oneoversamplerate = 0.;
	outlet_new(&x->x_obj, &s_signal);       
	return (x);
}


static t_int *phasor2_perform(t_int *w)
{
	t_phasor2 *x = (t_phasor2 *)(w[1]);
	t_sample *freq = (t_sample *)(w[2]);
	t_sample *phaseout = (t_sample *)(w[3]);
	int vecsize = (int)(w[4]); 
	
	t_float oneoversamplerate = x->oneoversamplerate;
	double phase = x->phase;
	double temp;
	int intphase;
	
	while (vecsize--)
	{
		// wrap phase within 0-1, but first check if wrap is needed (mostly it is not)
		if(phase>=1.)
		{
			intphase = (int)phase;
			phase -= intphase;
		}
		
		else if(phase<0.) 
		{
			intphase = (int)phase;
			intphase--;
			phase -= intphase;
		}
		
		temp = *freq++;
		*phaseout++ = phase;
		phase += temp * oneoversamplerate;
	}
	
	x->phase = phase;
	return (w+5);
}


static void phasor2_dsp(t_phasor2 *x, t_signal **sp)
{
	x->oneoversamplerate= 1. / sp[0]->s_sr;
	dsp_add(phasor2_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}


static void phasor2_ft1(t_phasor2 *x, t_floatarg f)
{
	x->phase = (double)f;
}


void phasor2_tilde_setup(void)
{
	phasor2_class = class_new(gensym("phasor2~"),
		(t_newmethod)phasor2_new, 0,
		sizeof(t_phasor2), 0, A_DEFFLOAT, 0);
	CLASS_MAINSIGNALIN(phasor2_class, t_phasor2, f);
	class_addmethod(phasor2_class, (t_method)phasor2_dsp,
		gensym("dsp"), 0);
	class_addmethod(phasor2_class, (t_method)phasor2_ft1,
        gensym("ft1"), A_FLOAT, 0);
}


/* ------------------------------------------------------- cos2~ ------------------------------------------------------- */


static t_class *cos2_class,*scalarosc_class;

typedef struct
	{
		t_object x_obj;
		t_float f;			// scalar frequency 
	} t_cos2;


static void *cos2_new(t_floatarg f)
{
	t_cos2 *x = (t_cos2 *)pd_new(cos2_class);
	x->f = f;
	outlet_new(&x->x_obj, &s_signal);       
	return (x);
}


static t_int *cos2_perform(t_int *w)
{
	t_cos2 *x = (t_cos2 *)(w[1]);
	t_sample *phase = (t_sample *)(w[2]);
	t_sample *cosine = (t_sample *)(w[3]);
	int vecsize = (int)(w[4]); 
	
	t_float *costab = cos_table2;
	double temp, cyclephase, tabphase;
	double fraction;
	int intphase, index;
	
	while (vecsize--)
	{
		cyclephase = *phase++;
		
		// wrap phase within 0-1, but first check if wrap is needed
		if(cyclephase>=1.)
		{
			intphase = (int)cyclephase;
			cyclephase -= intphase;
		}
		
		else if(cyclephase<0.)
		{
			intphase = (int)cyclephase;
			intphase--;
			cyclephase -= intphase;
		}
		
		// linear interpolation
		tabphase = cyclephase * COSTABSIZE;
		index = (int)tabphase;
		fraction = tabphase - index;
		*cosine++ = costab[index] + fraction * (costab[index+1] - costab[index]);
	}
	
	return (w+5);
}


static void cos2_dsp(t_cos2 *x, t_signal **sp)
{
	dsp_add(cos2_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}


void cos2_tilde_setup(void)
{
	cos2_class = class_new(gensym("cos2~"),
						   (t_newmethod)cos2_new, 0,
						   sizeof(t_cos2), 0, A_DEFFLOAT, 0);
	CLASS_MAINSIGNALIN(cos2_class, t_cos2, f);
	class_addmethod(cos2_class, (t_method)cos2_dsp,
					gensym("dsp"), 0);
}



/* ----------------------------------------------------------- osc2~ --------------------------------------------------------- */


static t_class *osc2_class,*scalarosc_class;

typedef struct
	{
		t_object x_obj;
		double phase;
		double oneoversamplerate;
		t_float f;						// scalar frequency 
	} t_osc2;


static void *osc2_new(t_floatarg f)
{
	t_osc2 *x = (t_osc2 *)pd_new(osc2_class);
	inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_float, gensym("ft1"));
	x->f = f;
	x->phase = 0.;
	x->oneoversamplerate = 0.;
	outlet_new(&x->x_obj, &s_signal);       
	return (x);
}


static t_int *osc2_perform(t_int *w)
{
	t_osc2 *x = (t_osc2 *)(w[1]);
	t_sample *freq = (t_sample *)(w[2]);
	t_sample *cosine = (t_sample *)(w[3]);
	int vecsize = (int)(w[4]); 
	
	t_float *costab = cos_table2;
	double oneoversamplerate = x->oneoversamplerate;
	double cyclephase = x->phase;
	double temp, tabphase, fraction;
	int intphase, index;
	
	while (vecsize--)
	{
		// wrap phase within 0-1, but first check if wrap is needed (mostly it is not)
		if(cyclephase>=1.)
		{
			intphase = (int)cyclephase;
			cyclephase -= intphase;
		}
		
		else if(cyclephase<0.)
		{
			intphase = (int)cyclephase;
			intphase--;
			cyclephase -= intphase;
		}
		
		temp = *freq++;
		
		// linear interpolation
		tabphase = cyclephase * COSTABSIZE;
		index = (int)tabphase;
		fraction = tabphase - index;
		*cosine++ = costab[index] + fraction * (costab[index+1] - costab[index]);
		
		cyclephase += temp * oneoversamplerate;
	}
	
	x->phase = cyclephase;
	return (w+5);
}


static void osc2_dsp(t_osc2 *x, t_signal **sp)
{
	x->oneoversamplerate = 1. / sp[0]->s_sr;
	dsp_add(osc2_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}


static void osc2_ft1(t_osc2 *x, t_float phase)
{
	x->phase = phase;
}


void osc2_tilde_setup(void)
{
	osc2_class = class_new(gensym("osc2~"),
						   (t_newmethod)osc2_new, 0,
						   sizeof(t_osc2), 0, A_DEFFLOAT, 0);
	CLASS_MAINSIGNALIN(osc2_class, t_osc2, f);
	class_addmethod(osc2_class, (t_method)osc2_dsp,
					gensym("dsp"), 0);
	class_addmethod(osc2_class, (t_method)osc2_ft1,
					gensym("ft1"), A_FLOAT, 0);
}


/* ------------------------------------------------- vcf2~ - 2-pole bandpass filter -------------------------------------- */

typedef struct vcfctl
	{
		t_float real;
		t_float im;
		t_float q;
		t_float oneoversamplerate;
	} t_vcfctl;


typedef struct vcf2
	{
		t_object x_obj;
		t_vcfctl cspace;
		t_vcfctl *ctl;
		t_float f;
	} t_vcf2;

t_class *vcf2_class;


static void *vcf2_new(t_floatarg q)
{
	t_vcf2 *x = (t_vcf2 *)pd_new(vcf2_class);
	inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_signal, &s_signal);
	inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("ft1"));
	outlet_new(&x->x_obj, gensym("signal"));
	outlet_new(&x->x_obj, gensym("signal"));
	x->ctl = &x->cspace;
	x->cspace.real = 0;
	x->cspace.im = 0;
	x->cspace.q = q;
	x->cspace.oneoversamplerate = 0;
	x->f = 0;
	return (x);
}


static void vcf2_ft1(t_vcf2 *x, t_floatarg q)
{
	x->ctl->q = (q> 0. ? q : 0.);
}


static t_int *vcf2_perform(t_int *w)
{
	t_sample *in1 = (t_sample *)(w[1]);
	t_sample *in2 = (t_sample *)(w[2]);
	t_sample *out1 = (t_sample *)(w[3]);
	t_sample *out2 = (t_sample *)(w[4]);
	t_vcfctl *c = (t_vcfctl *)(w[5]);
	int vectorsize = (t_int)(w[6]);
	
	t_float real = c->real, real2;
	t_float im = c->im;
	t_float q = c->q;
	
	t_float qinv = 0.; 
	if(q > 0.) qinv = 1. / q;
	t_float ampcorrect = 2. - 2. / (q + 2.);
	t_float *costab = cos_table2;
	t_float freqnorm = c->oneoversamplerate * TWOPI;
	t_float tabnorm =  c->oneoversamplerate * COSTABSIZE;
	t_float realcoef, imcoef, fraction, insamp;
	t_float tabphase, centerfreq, radius, oneminusr;
	int index;
	
	while(vectorsize--)
	{
		centerfreq = *in2++;						
		if(centerfreq < 0.) centerfreq = 0.;						// constrict lower frequency
		tabphase = centerfreq;
		centerfreq *= freqnorm;								// frequency normalized to interval 0 - 2pi
		radius = 1. - centerfreq * qinv - EPSILON;				// subtract EPSILON, just in case (centerfreq * qinv) is too small
		if ((radius < 0.) || (qinv == 0.)) radius = 0.;
		oneminusr = 1.0 - radius;
		
		tabphase *= tabnorm;
		if(tabphase >= COSTABSIZE) tabphase = COSTABSIZE;	// constrict upper frequency
		index = (int)tabphase;
		fraction = tabphase - index;
		
		realcoef = radius * (costab[index] + fraction * (costab[index+1] - costab[index]));
		
		index -= COSTABSIZE>>2;
		index &= COSTABSIZE-1;
		
		imcoef = radius * (costab[index] + fraction * (costab[index+1] - costab[index]));
		
		insamp = *in1++;
		real2 = real;
		*out1++ = real = ampcorrect * oneminusr * insamp + realcoef * real2 - imcoef * im;
		*out2++ = im = imcoef * real2 + realcoef * im;
	}
	
	if (PD_BIGORSMALL(real)) real = 0.;
	if (PD_BIGORSMALL(im)) im = 0.;
	c->real = real;
	c->im = im;
	return (w+7);
}


static void vcf2_dsp(t_vcf2 *x, t_signal **sp)
{
	x->ctl->oneoversamplerate = 1. / sp[0]->s_sr;
	dsp_add(vcf2_perform, 6, sp[0]->s_vec, sp[1]->s_vec, 
			sp[2]->s_vec, sp[3]->s_vec, x->ctl, sp[0]->s_n);
}


void vcf2_tilde_setup(void)
{
	vcf2_class = class_new(gensym("vcf2~"), (t_newmethod)vcf2_new, 0,
			sizeof(t_vcf2), 0, A_DEFFLOAT, 0);
	CLASS_MAINSIGNALIN(vcf2_class, t_vcf2, f);
	class_addmethod(vcf2_class, (t_method)vcf2_dsp, gensym("dsp"), 0);
	class_addmethod(vcf2_class, (t_method)vcf2_ft1,
			gensym("ft1"), A_FLOAT, 0);

}

/* ---------------------------------------------------- bigorsmall_pun~ ------------------------------------------------ */
#if PD_FLOAT_PRECISION == 32
#define BIGORSMALL_PUN(f) ((((*(unsigned int*)&(f))&0x60000000)==0) || \
	(((*(unsigned int*)&(f))&0x60000000)==0x60000000))
#elif PD_FLOAT_PRECISION == 64

// this does not work, it always resolves to 'false':
#define BIGORSMALL_PUN(f) ((((*(unsigned long long*)&(f))&0x6000000000000000ULL)==0ULL) || \
	(((*(unsigned long long*)&(f))&0x6000000000000000ULL)==0x6000000000000000ULL))
#endif

static t_class *bigorsmall_pun_class;

typedef struct
{
	t_object x_obj;
	t_float f;
}t_bigorsmall_pun;
	

static void *bigorsmall_pun_new(t_floatarg f)
{
	t_bigorsmall_pun *x =(t_bigorsmall_pun *)pd_new(bigorsmall_pun_class);
	x->f = f;
	outlet_new(&x->x_obj, &s_signal);       
	return (x);
}


static t_int *bigorsmall_pun_perform(t_int *w)
{
	t_bigorsmall_pun *x = (t_bigorsmall_pun *)(w[1]);
	t_sample *in = (t_sample *)(w[2]);
	t_sample *out = (t_sample *)(w[3]);
	int vecsize = (int)(w[4]); 
	
	t_sample temp;
	
	while(vecsize--)
	{
		temp = *in++;
		if(BIGORSMALL_PUN(temp)) temp = 0.;
		*out++ = temp;
	}
	
	return(w+5);
}


static void bigorsmall_pun_dsp(t_bigorsmall_pun *x, t_signal **sp)
{
	dsp_add(bigorsmall_pun_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}


void bigorsmall_pun_tilde_setup(void)
{
	bigorsmall_pun_class = class_new(gensym("bigorsmall_pun~"),
		(t_newmethod)bigorsmall_pun_new, 0, sizeof(t_bigorsmall_pun),
		0, A_DEFFLOAT, 0);
	CLASS_MAINSIGNALIN(bigorsmall_pun_class, t_bigorsmall_pun, f);
	class_addmethod(bigorsmall_pun_class, (t_method)bigorsmall_pun_dsp,
		gensym("dsp"), 0);
}
	

/* ---------------------------------------------------- bigorsmall_nopun~ ------------------------------------------------ */
#define BIGVALUE pow(2., 64.)					 
#define SMALLVALUE pow(2., -64.)				 
#define BIGORSMALL_NOPUN(f) ((fabs(f) <= SMALLVALUE) || (fabs(f) >= BIGVALUE) || isnan(f))


static t_class *bigorsmall_nopun_class;

typedef struct
	{
		t_object x_obj;
		t_float f;
	}t_bigorsmall_nopun;


static void *bigorsmall_nopun_new(t_floatarg f)
{
	t_bigorsmall_nopun *x =(t_bigorsmall_nopun *)pd_new(bigorsmall_nopun_class);
	x->f = f;
	outlet_new(&x->x_obj, &s_signal);       
	return (x);
}


static t_int *bigorsmall_nopun_perform(t_int *w)
{
	t_bigorsmall_nopun *x = (t_bigorsmall_nopun *)(w[1]);
	t_sample *in = (t_sample *)(w[2]);
	t_sample *out = (t_sample *)(w[3]);
	int vecsize = (int)(w[4]); 
	
	t_sample temp;
	
	while(vecsize--)
	{
		temp = *in++;
		if(BIGORSMALL_NOPUN(temp)) temp = 0.;
		*out++ = temp;
	}
	
	return(w+5);
}


static void bigorsmall_nopun_dsp(t_bigorsmall_nopun *x, t_signal **sp)
{
	dsp_add(bigorsmall_nopun_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}


void bigorsmall_nopun_tilde_setup(void)
{
	bigorsmall_nopun_class = class_new(gensym("bigorsmall_nopun~"),
		(t_newmethod)bigorsmall_nopun_new, 0, sizeof(t_bigorsmall_nopun),
		0, A_DEFFLOAT, 0);
	CLASS_MAINSIGNALIN(bigorsmall_nopun_class, t_bigorsmall_nopun, f);
	class_addmethod(bigorsmall_nopun_class, (t_method)bigorsmall_nopun_dsp,
					gensym("dsp"), 0);
}




/* ---------------------------------------------------- global setup routine ------------------------------------------------ */
// When building d_osc_rewrite.c into the pd executable, this global setup routine is used. 
// Be sure to call the setup function from m_conf.c.

#if 1
void d_doubletest_setup(void)
{
	cos_maketable2();
	phasor2_tilde_setup();
	cos2_tilde_setup();
	osc2_tilde_setup();
	vcf2_tilde_setup();
	bigorsmall_pun_tilde_setup();
	bigorsmall_nopun_tilde_setup();

	post("d_doubletest: load cos_table2");
	post("d_doubletest: setup phasor2~, osc2~, cos2~, vcf2~");
	post("d_doubletest: setup bigorsmall_pun~ and bigorsmall_nopun~");
	post("PD_FLOATPRECISION = %d bits", sizeof(t_float)*8);
}
#endif


/* When d_osc_rewrite.c is compiled as a separate executable file, 
not part of pd, the global setup routine must be able to create an object.
Enable the alternative code below.
 
An object box with d_doubletest must be created on a Pd patch   
in order to create cos_table2 and set up phasor2~, cos2~, osc2~, vcf2~, 
bigorsmall_pun~ and bigorsmal_nopun*/

#if 0
typedef struct
	{	
		t_object x_obj;
	}t_d_doubletest;

static t_class *d_doubletest_class;


static void *d_doubletest_new(void)
{
	t_d_doubletest *x = (t_d_doubletest*)pd_new(d_doubletest_class);
	return (x);
}


void d_doubletest_setup(void)
{
	d_osc_rewrite_class = class_new(gensym("d_doubletest"), 
		(t_newmethod)d_doubletest_new, 0, sizeof(t_d_doubletest), CLASS_DEFAULT, 0);
	
	if(!cos_table2) cos_maketable2();
	if(!phasor2_class) phasor2_tilde_setup();
	if(!cos2_class) cos2_tilde_setup();
	if(!osc2_class) osc2_tilde_setup();
	if(!vcf2_class) vcf2_tilde_setup();
	if(!biquad2~class) biquad2_tilde_setup();
	post("d_doubletest: load cos_table2");
	post("d_doubletest: setup of phasor2~, osc2~, cos2~, vcf2~");
	post("d_doubletest: setup bigorsmall_pun~ and bigorsmall_nopun~");
	post("PD_FLOATPRECISION = %d bits", sizeof(t_float)*8);
	//post("d_doubletest: (1. - EPSILON)*10. = %f", (1. - EPSILON)*10.);
}
#endif



